/*
 * fontwork.c --- fontwork functions
 * 
 * Copyright (c) 2019, 2022 Mariano Alvarez Fernandez (malfer@telefonica.net)
 *
 * This file is part of FEdit, a visual font editor for the MGRX font format
 *
 * FEdit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FEdit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FEdit. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <grgui.h>
#include "fedit.h"

static FontWork *_FontWorkNew(int maxglyphs, int boxwidth, int boxheight);
static void translateglyphbmp(FontWork *dfwk, FontWork *sfwk, int sglyph, int dglyph,
                              int top, int left, int right, int bottom, int clear);

FontWork *FontWorkCreate(GrFont *font, int maxglyphs)
{
    FontWork *fwk;
    int i, firstglyph, lastglyph;

    if (maxglyphs == 0) maxglyphs = font->h.minchar + font->h.numchars;

    firstglyph = font->h.minchar;
    lastglyph = font->h.minchar + font->h.numchars - 1;
    //if (lastglyph >= maxglyphs) maxglyphs = lastglyph + 1;
    //printf("previo %d %d %d\n", maxglyphs, firstglyph, lastglyph);
    if ((firstglyph >= maxglyphs) || (lastglyph >= maxglyphs)) return NULL;

    //printf("antes %d %d %d\n", maxglyphs, font->maxwidth, font->h.height);
    fwk = _FontWorkNew(maxglyphs, font->maxwidth, font->h.height);
    if (fwk == NULL) return NULL;
    
    fwk->h = font->h;

    for (i=firstglyph; i<=lastglyph; i++) {
        FontWorkLoadGlyph(fwk, font, i);
    }

    // if no memory == NULL, so don worry now
    fwk->h.name = strdup(font->h.name);
    fwk->h.family = strdup(font->h.family);

    fwk->glyphselec = 0;
    fwk->xcent = 0;
    fwk->ycent = 0;
    fwk->yorg = 0;
    fwk->glyphxrow = 1;
    fwk->firstline = 0;
    fwk->numlines = 1;
    fwk->screenlines = 1;

    fwk->bgcolor = GrBlack();
    fwk->fgcolor = GrWhite();
    fwk->bgincolor = GrBlack();
    fwk->selectcolor = GrWhite();
    fwk->noinitcolor = GrWhite();

    return fwk;
}

int FontWorkAppend(FontWork *dfwk, GrFont *sf, int fsglyph, int lsglyph,
                   int fdglyph, int xpivot, int ypivot)
{
    FontWork *sfwk, *auxfwk;
    int top, left, right, bottom;
    int si, di;

    if ((fsglyph < sf->h.minchar) ||
        (lsglyph >= sf->h.minchar + sf->h.numchars)) return 0;

    if (!FontWorkCheckValidRange(dfwk, fdglyph, fdglyph+lsglyph-fsglyph+1)) return 0;

    sfwk = FontWorkCreate(sf, 0);
    if (sfwk == NULL) return 0;
    
    if (dfwk->h.proportional && !sfwk->h.proportional) {
        sfwk->h.proportional = 1;
    }

    if (!dfwk->h.proportional && sfwk->h.proportional) {
        auxfwk = FontWorkConvToFixed(sfwk);
        FontWorkDestroy(sfwk);
        if (auxfwk == NULL) return 0;
        sfwk = auxfwk;
    }

    top = ypivot;
    left = xpivot;
    bottom = dfwk->h.height - ypivot - sf->h.height;
    right = dfwk->maxwidth - xpivot -sf->maxwidth;

    di = fdglyph;
    for (si=fsglyph; si<=lsglyph; si++) {
        translateglyphbmp(dfwk, sfwk, si, di, top, left, right, bottom, 1);
        di++;
    }

    FontWorkDestroy(sfwk);
    return 1;
}

FontWork *FontWorkResizeBox(FontWork *fwk, int top, int left,
                             int right, int bottom)
{
    FontWork *fwk2;
    int newmaxwidth, newheight;
    int i;

    newmaxwidth = fwk->maxwidth + left + right;
    newheight = fwk->h.height + top + bottom;
    // we impose a minimal box of 4x4
    if (newmaxwidth < 4) return NULL;
    if (newheight < 4) return NULL;

    fwk2 = _FontWorkNew(fwk->maxglyphs, newmaxwidth, newheight);
    if (fwk2 == NULL) return NULL;

    fwk2->h = fwk->h;
    fwk2->h.height = newheight;
    fwk2->h.baseline += top;
    if (fwk2->h.baseline < 0) fwk2->h.baseline = 0;
    if (fwk2->h.baseline >= fwk2->h.height) fwk2->h.baseline = fwk2->h.height - 1;

    for (i=0; i<fwk->maxglyphs; i++) {
        translateglyphbmp(fwk2, fwk, i, i, top, left, right, bottom, 0);
    }

    FontWorkRecalMeanWidth(fwk2);

    fwk2->h.name = strdup(fwk->h.name);
    fwk2->h.family = strdup(fwk->h.family);

    fwk2->glyphselec = 0;
    fwk2->xcent = 0;
    fwk2->ycent = 0;
    fwk2->yorg = 0;
    fwk2->glyphxrow = 1;
    fwk2->firstline = 0;
    fwk2->numlines = 1;
    fwk2->screenlines = 1;

    fwk2->bgcolor = fwk->bgcolor;
    fwk2->fgcolor = fwk->fgcolor;
    fwk2->bgincolor = fwk->bgincolor;
    fwk2->selectcolor = fwk->selectcolor;
    fwk2->noinitcolor = fwk->noinitcolor;

    if (fwk->t) FontWorkSetDims(fwk2, fwk->t);

    return fwk2;
}

FontWork *FontWorkConvToFixed(FontWork *fwk)
{
    FontWork *fwk2;
    int i, bmove;

    if (!fwk->h.proportional) return NULL;

    fwk2 = _FontWorkNew(fwk->maxglyphs, fwk->maxwidth, fwk->h.height);
    if (fwk2 == NULL) return NULL;

    fwk2->h = fwk->h;
    fwk2->h.proportional = 0;

    for (i=0; i<fwk->maxglyphs; i++) {
        bmove = ((fwk->maxwidth - fwk->gly[i].width) + 1) / 2;
        translateglyphbmp(fwk2, fwk, i, i, 0, bmove, -bmove, 0, 0);
    }

    FontWorkRecalMeanWidth(fwk2);

    fwk2->h.name = strdup(fwk->h.name);
    fwk2->h.family = strdup(fwk->h.family);

    fwk2->glyphselec = 0;
    fwk2->xcent = 0;
    fwk2->ycent = 0;
    fwk2->yorg = 0;
    fwk2->glyphxrow = 1;
    fwk2->firstline = 0;
    fwk2->numlines = 1;
    fwk2->screenlines = 1;

    fwk2->bgcolor = fwk->bgcolor;
    fwk2->fgcolor = fwk->fgcolor;
    fwk2->bgincolor = fwk->bgincolor;
    fwk2->selectcolor = fwk->selectcolor;
    fwk2->noinitcolor = fwk->noinitcolor;

    if (fwk->t) FontWorkSetDims(fwk2, fwk->t);

    return fwk2;
}

void FontWorkLoadGlyph(FontWork *fwk, GrFont *font, int glyph)
{
    char *bmp;
    
    if (glyph < 0 || glyph >= fwk->maxglyphs) return;
    if (!GrFontCharPresent(font, glyph)) return;

    fwk->gly[glyph].inited = 1;
    fwk->gly[glyph].width = GrFontCharWidth(font, glyph);
    fwk->gly[glyph].bmprowsize = GrFontCharBmpRowSize(font, glyph);
    bmp = GrFontCharBitmap(font, glyph);
    memcpy(fwk->gly[glyph].bmp, bmp, GrFontCharBitmapSize(font, glyph));
}

void FontWorkSetDims(FontWork *fwk, GUITile *t)
{
    int tw, th, fwkw, fwkh;

    fwk->t = t;
    tw = t->p->cl->gc_xmax + 1;
    th = t->p->cl->gc_ymax + 1;

    fwk->glyphxrow = tw / (fwk->maxwidth + 4);
    fwkw = fwk->glyphxrow * (fwk->maxwidth + 4);
    fwk->xcent = (tw - fwkw) / 2;
    fwk->yorg = 0;
    
    fwk->numlines = ((fwk->maxglyphs + fwk->glyphxrow - 1) / fwk->glyphxrow);
    fwk->screenlines = th / (fwk->h.height + 4);
    if (fwk->screenlines >= fwk->numlines) {
        fwk->screenlines = fwk->numlines;
    }
    fwkh = fwk->screenlines * (fwk->h.height + 4);
    fwk->ycent = (th - fwkh) / 2;

    GUIScrollbarSetLimits(t->p->vscb, 0, fwk->numlines-1,
                          0, fwk->screenlines-1);
    
}

void FontWorkSetColors(FontWork *fwk, GrColor bg, GrColor fg, GrColor bgin,
                       GrColor select, GrColor noinit)
{
    fwk->bgcolor = bg;
    fwk->fgcolor = fg;
    fwk->bgincolor = bgin;
    fwk->selectcolor = select;
    fwk->noinitcolor = noinit;
}

void FontWorkPaintGlyphXY(FontWork *fwk, int glyph, int x, int y, int selec, int blit)
{
    static GrLineOption lopt = {0, 1, 2, (unsigned char *)"\1\1"};
    GrContext grcaux;
    GrColor caux;
    char *p;
    int i, j;
    int byte, bit;
    int mdx, mdy;

    GrSaveContext(&grcaux);
    GrSetContext(fwk->wctx);

    if (glyph >= fwk->h.minchar && glyph < fwk->h.numchars + fwk->h.minchar)
        GrClearContext(fwk->bgincolor);
    else
        GrClearContext(fwk->bgcolor);

    if ((glyph >= 0) && (glyph < fwk->maxglyphs)) {
        if (selec) {
            //caux = (fwk->gly[glyph].inited) ? fwk->selectcolor : fwk->noinitcolor;
            caux = fwk->selectcolor;
            GrBox(0, 0, fwk->maxwidth+3, fwk->h.height+3, caux);
        }
 
        if (fwk->gly[glyph].inited) {
            for (i=0; i<fwk->h.height; i++) {
                p = fwk->gly[glyph].bmp + (fwk->gly[glyph].bmprowsize * i);
                for (j=0; j<fwk->gly[glyph].width; j++) {
                    byte = j / 8;
                    bit = 0x80 >> (j % 8);
                    if (p[byte] & bit)
                        GrPlot(j+2, i+2, fwk->fgcolor);
                }
            }
        } else {
            lopt.lno_color = fwk->noinitcolor;
            //GrLine(3, 3, fwk->maxwidth, fwk->h.height, fwk->noinitcolor);
            //GrLine(3, fwk->h.height, fwk->maxwidth, 3, fwk->noinitcolor);
            GrCustomLine(3, 3, fwk->maxwidth, fwk->h.height, &lopt);
            GrCustomLine(3, fwk->h.height, fwk->maxwidth, 3, &lopt);
        }
    } else { // ghost glyph
        mdx = (fwk->maxwidth + 4) / 2;
        mdy = (fwk->h.height + 4) / 2;
        GrFilledBox(mdx-2, mdy-2, mdx+2, mdy+2 , fwk->noinitcolor);
    }

    GrSetContext(&grcaux);
    GrBitBlt(NULL, x, y, fwk->wctx, 0, 0,
             fwk->maxwidth+3, fwk->h.height+3, GrWRITE);
    if (blit) {
        GUIDBCurCtxBltRectToScreen(x, y, x+fwk->maxwidth+3, y+fwk->h.height+3);
    }
}

void FontWorkPaintGlyph(FontWork *fwk, int glyph, int selec, int blit)
{
    int r, c, x, y;

    r = glyph / fwk->glyphxrow;
    c = glyph % fwk->glyphxrow;
    x = c * (fwk->maxwidth + 4) + fwk->xcent;
    y = r * (fwk->h.height + 4) + fwk->yorg + fwk->ycent;
    FontWorkPaintGlyphXY(fwk, glyph, x, y, selec, blit);
}

void FontWorkPaint(FontWork *fwk)
{
    int i, r, c;
    int minglyph, maxglyph;

    minglyph = fwk->firstline * fwk->glyphxrow;
    maxglyph = (fwk->firstline + fwk->screenlines) * fwk->glyphxrow;
    if (maxglyph > fwk->maxglyphs) maxglyph = fwk->maxglyphs;

    for (i=minglyph; i<maxglyph; i++)
        FontWorkPaintGlyph(fwk, i, (i==fwk->glyphselec), 0);

    // ghost glyphs to complete the line
    r = (maxglyph - 1) / fwk->glyphxrow;
    if (r == fwk->numlines - 1) {
        c = (maxglyph - 1) % fwk->glyphxrow;
        if (c < fwk->glyphxrow - 1) {
            for (i=0; i < fwk->glyphxrow-c-1; i++) {
                FontWorkPaintGlyph(fwk, maxglyph+i, 0, 0);
            }
        }
    }
    GUIDBCurCtxBltRectToScreen(0, 0, GrMaxX(), GrMaxY());
}

void FontWorkSetSelect(FontWork *fwk, int glyph, int paint)
{
    int oldglyph;
    int line, diffline;

    if (glyph < 0 || glyph >= fwk->maxglyphs) return;

    oldglyph = fwk->glyphselec;
    fwk->glyphselec = glyph;

    //if (paint) {
        line = glyph / fwk->glyphxrow;
        diffline = line - fwk->firstline;
        if (diffline < 0) {
            fwk->firstline += diffline;
            fwk->yorg -= diffline * (fwk->h.height + 4);
            if (paint) FontWorkPaint(fwk);
            GUIScrollbarSetLimits(fwk->t->p->vscb, 0, fwk->numlines-1,
                fwk->firstline, fwk->firstline+fwk->screenlines-1);
            if (paint) GUIPanelRePaintScb(fwk->t->p);
        } else if (diffline > fwk->screenlines-1) {
            diffline -= (fwk->screenlines - 1);
            fwk->firstline += diffline;
            fwk->yorg -= diffline * (fwk->h.height + 4);
            if (paint) FontWorkPaint(fwk);
            GUIScrollbarSetLimits(fwk->t->p->vscb, 0, fwk->numlines-1,
                fwk->firstline, fwk->firstline+fwk->screenlines-1);
            if (paint) GUIPanelRePaintScb(fwk->t->p);
        } else {
            if (paint) FontWorkPaintGlyph(fwk, oldglyph, 0, 1);
            if (paint) FontWorkPaintGlyph(fwk, glyph, 1, 1);
        }
    //}
}

void FontWorkMoveSelect(FontWork *fwk, int row, int col, int paint)
{
    int r, c, newglyph;

    r = (fwk->glyphselec / fwk->glyphxrow) + row;
    c = (fwk->glyphselec % fwk->glyphxrow) + col;
    newglyph = (r * fwk->glyphxrow) + c;
    FontWorkSetSelect(fwk, newglyph, paint);
}

void FontWorkMovePagSelect(FontWork *fwk, int direc, int paint)
{
    int pag, row, col, newrow, newglyph;

    pag = fwk->screenlines - 1;
    row = fwk->glyphselec / fwk->glyphxrow;
    col = fwk->glyphselec % fwk->glyphxrow;
    newrow = row + (direc * pag);
    if (newrow < 0) newrow = 0;
    if (newrow >= fwk->numlines) newrow = fwk->numlines - 1;
    newglyph = (newrow * fwk->glyphxrow) + col;
    if (newglyph >= fwk->maxglyphs) newglyph = fwk->maxglyphs - 1;
    FontWorkSetSelect(fwk, newglyph, paint);
}

void FontWorkScrollSelect(FontWork *fwk, int newfirst, int paint)
{
    int diffrow, newrow, row, col, newglyph;

    if ((newfirst < 0)) newfirst = 0;
    if (newfirst > fwk->numlines - fwk->screenlines)
        newfirst = fwk->numlines - fwk->screenlines;
    if (newfirst == fwk->firstline) return;

    diffrow = newfirst - fwk->firstline;
    row = fwk->glyphselec / fwk->glyphxrow;
    col = fwk->glyphselec % fwk->glyphxrow;
    newrow = row + diffrow;
    if (newrow < 0) newrow = 0;
    if (newrow >= fwk->numlines) newrow = fwk->numlines - 1;
    newglyph = (newrow * fwk->glyphxrow) + col;
    if (newglyph >= fwk->maxglyphs) newglyph = fwk->maxglyphs - 1;
    fwk->glyphselec = newglyph;
    fwk->firstline = newfirst;
    fwk->yorg = -newfirst * (fwk->h.height + 4);
    if (paint) FontWorkPaint(fwk);
    GUIScrollbarSetLimits(fwk->t->p->vscb, 0, fwk->numlines-1,
        fwk->firstline, fwk->firstline+fwk->screenlines-1);
    if (paint) GUIPanelRePaintScb(fwk->t->p);
}

int FontWorkCalcGlyphByGCoord(FontWork *fwk, int gx, int gy, int *glyph)
{
    int r, c;

    if ((gx < fwk->xcent) || (gx > fwk->t->p->cl->gc_xmax - fwk->xcent)) return 0;
    if ((gy < fwk->ycent) || (gy > fwk->t->p->cl->gc_ymax - fwk->ycent)) return 0;

    c = (gx - fwk->xcent) / (fwk->maxwidth + 4);
    r = (gy - fwk->yorg - fwk->ycent) / (fwk->h.height + 4);
    *glyph = (r * fwk->glyphxrow) + c;
    if (*glyph < 0 || *glyph >= fwk->maxglyphs) return 0;
    return 1;
}

void FontWorkGetRealSelectedXYPos(FontWork *fwk, int *x, int *y)
{
    int r, c;

    r = fwk->glyphselec / fwk->glyphxrow;
    c = fwk->glyphselec % fwk->glyphxrow;
    //*x = c * (fwk->maxwidth + 4) + fwk->xcent;
    //*y = (r + 1) * (fwk->h.height + 4) + fwk->yorg + fwk->ycent;
    *x = (c + 1) * (fwk->maxwidth + 4) + fwk->xcent;
    *y = r * (fwk->h.height + 4) + fwk->yorg + fwk->ycent;
    *x += fwk->t->p->cl->gc_xoffset;
    *y += fwk->t->p->cl->gc_yoffset;
}

void FontWorkCopyGlyph(FontWork *fwk, int sglyph, int dglyph)
{
    int byxgly;
    
    if (sglyph < 0 || sglyph >= fwk->maxglyphs) return;
    if (dglyph < 0 || dglyph >= fwk->maxglyphs) return;
    if (sglyph == dglyph) return;

    fwk->gly[dglyph].inited = fwk->gly[sglyph].inited;
    fwk->gly[dglyph].width = fwk->gly[sglyph].width;
    fwk->gly[dglyph].bmprowsize = fwk->gly[sglyph].bmprowsize;
    byxgly = ((fwk->maxwidth + 7) / 8) * fwk->h.height;
    memcpy(fwk->gly[dglyph].bmp, fwk->gly[sglyph].bmp, byxgly);
}

GrFont *FontWorkGenFont(FontWork *fwk)
{
    GrFont *f;
    int nb, i, index;
    char * pf;

    FontWorkRecalMeanWidth(fwk);

    nb = sizeof(GrFont) + ((fwk->h.numchars - 1) * sizeof(GrFontChrInfo));
    f = malloc(nb);
    if (f == NULL) goto error;
    memset(f, 0, nb);
    f->h = fwk->h;
    f->h.preloaded = 0;
    f->h.name = strdup(fwk->h.name);
    f->h.family = strdup(fwk->h.family);
    if ((f->h.name == NULL) || (f->h.family == NULL)) goto error;
    f->minwidth = 0x7fff;
    f->maxwidth = 0;
    f->h.sparse = 0;

    nb = 0;
    for (i=0; i<f->h.numchars; i++) {
        index = i + f->h.minchar;
        //if (!fwk->gly[index].inited) goto error;
        if (!fwk->gly[index].inited) {
            f->chrinfo[i].width  = 0;
            f->chrinfo[i].offset = nb;
            f->h.sparse = 1;
        } else {
            f->chrinfo[i].width  = fwk->gly[index].width;
            f->chrinfo[i].offset = nb;
            nb += fwk->gly[index].bmprowsize * f->h.height;
            if (f->minwidth > f->chrinfo[i].width) f->minwidth = f->chrinfo[i].width;
            if (f->maxwidth < f->chrinfo[i].width) f->maxwidth = f->chrinfo[i].width;
        }
    }

    f->bitmap = malloc(nb);
    if (f->bitmap == NULL) goto error;
    nb = 0;
    for (i=0; i<f->h.numchars; i++) {
        index = i + f->h.minchar;
        if (!fwk->gly[index].inited) continue;
        pf = f->bitmap + f->chrinfo[i].offset;
        memcpy(pf, fwk->gly[index].bmp, fwk->gly[index].bmprowsize*f->h.height);
    }

    return f;

    error:
    if (f != NULL) {
        if (f->h.name) free(f->h.name);
        if (f->h.family) free(f->h.family);
        if (f->bitmap) free(f->bitmap);
        free(f);
    }
    return NULL;
}

int FontWorkCheckUninitedInRange(FontWork *fwk, int fglyph, int lglyph)
{
    int i;

    for (i=fglyph; i<lglyph; i++) {
        if (!fwk->gly[i].inited) return 1;
    }

    return 0;
}

int FontWorkCheckUninitedInFontRange(FontWork *fwk)
{
    int i, fglyph, lglyph;

    fglyph = fwk->h.minchar;
    lglyph = fwk->h.numchars + fwk->h.minchar - 1;
    for (i=fglyph; i<=lglyph; i++) {
        if (!fwk->gly[i].inited) return 1;
    }

    return 0;
}

int FontWorkCheckInitedOutOfRange(FontWork *fwk, int fglyph, int lglyph)
{
    int i;

    for (i=0; i<fglyph; i++) {
        if (fwk->gly[i].inited) return 1;
    }
    for (i=lglyph+1; i<fwk->maxglyphs; i++) {
        if (fwk->gly[i].inited) return 1;
    }

    return 0;
}

int FontWorkCheckValidRange(FontWork *fwk, int fglyph, int lglyph)
{
    if ((fglyph >= 0) &&
        (fglyph <= lglyph) &&
        (lglyph < fwk->maxglyphs)) return 1;

    return 0;
}

int FontWorkCheckGlyphOk(FontWork *fwk, int glyph)
{
    if ((glyph >= 0) && (glyph < fwk->maxglyphs)) return 1;

    return 0;
}

void FontWorkRecalMeanWidth(FontWork *fwk)
{
    int i, index;
    long sumwidth = 0;
    int nitems = 0;

    if (fwk->h.proportional) {
        for (i=0; i<fwk->h.numchars; i++) {
            index = i + fwk->h.minchar;
            if (fwk->gly[index].inited) {
                sumwidth += fwk->gly[index].width;
                nitems++;
            }
        }
        if (nitems > 0) {
            fwk->h.width = sumwidth / nitems;
        } else {
            fwk->h.width = fwk->maxwidth;
        }
    } else {
        fwk->h.width = fwk->maxwidth;
    }
}

void FontWorkDestroy(FontWork *fwk)
{
    if (fwk) {
        if (fwk->h.name) free(fwk->h.name);
        if (fwk->h.family) free(fwk->h.family);
        if (fwk->gly) free(fwk->gly);
        if (fwk->bmp) free(fwk->bmp);
        if (fwk->wctx) GrDestroyContext(fwk->wctx);
        free(fwk);
    }
}

static FontWork *_FontWorkNew(int maxglyphs, int boxwidth, int boxheight)
{
    FontWork *fwk;
    int byxgly;
    char *pbmp;
    int i;

    if (maxglyphs <= 0) return NULL;
    if (boxwidth <= 0) return NULL;
    if (boxheight <= 0) return NULL;

    fwk = (FontWork *)calloc(1, sizeof(FontWork));
    if (fwk == NULL) return NULL;

    fwk->maxglyphs = maxglyphs;
    fwk->maxwidth = boxwidth;
    fwk->h.height = boxheight;
    fwk->t = NULL;
    
    fwk->gly = (FontWorkGlyph *)calloc(fwk->maxglyphs, sizeof(FontWorkGlyph));
    if (fwk->gly == NULL) {
        free(fwk);
        return NULL;
    }

    byxgly = ((fwk->maxwidth + 7) / 8) * fwk->h.height;
    fwk->bmp = (char *)calloc(byxgly * fwk->maxglyphs, sizeof(char));
    if (fwk->bmp == NULL) {
        free(fwk->gly);
        free(fwk);
        return NULL;
    }

    pbmp = fwk->bmp;
    for (i=0; i<fwk->maxglyphs; i++) {
        fwk->gly[i].inited = 0;
        fwk->gly[i].bmp = pbmp;
        pbmp += byxgly;
    }

    fwk->wctx = GrCreateContext(fwk->maxwidth+4, fwk->h.height+4, NULL, NULL);
    if (fwk->wctx == NULL) {
        free(fwk->bmp);
        free(fwk->gly);
        free(fwk);
        return NULL;
    }

    return fwk;
}

static void translateglyphbmp(FontWork *dfwk, FontWork *sfwk, int sglyph, int dglyph,
                              int top, int left, int right, int bottom, int clear)
{
    int fsbit, lsbit, fdbit;
    int fsrow, lsrow, fdrow;
    int nbits, nrows, r, b;
    int sbyte, dbyte, sbit, dbit;
    int exceed;
    char *sp, *dp;
    int byxgly;

    if (!sfwk->gly[sglyph].inited) return;
    dfwk->gly[dglyph].inited = 1;
    // begin with some defaults
    if (dfwk->h.proportional) {
        dfwk->gly[dglyph].width = 1;
        dfwk->gly[dglyph].bmprowsize = 1;
    } else {
        dfwk->gly[dglyph].width = dfwk->maxwidth;
        dfwk->gly[dglyph].bmprowsize = (dfwk->maxwidth + 7) / 8;
    }

    fsbit = 0;
    if (left < 0) fsbit = -left;
    lsbit = sfwk->gly[sglyph].width - 1;
    if (right < 0) {
        exceed = -right - (sfwk->maxwidth - sfwk->gly[sglyph].width);
        if (exceed > 0)
            lsbit = sfwk->gly[sglyph].width - 1 - exceed;
    }
    nbits = lsbit - fsbit + 1;
    if (nbits <= 0) return;

    fdbit = 0;
    if (left > 0) fdbit = left;
    if (fdbit + nbits > dfwk->maxwidth) return;

    fsrow = 0;
    if (top < 0) fsrow = -top;
    lsrow = sfwk->h.height - 1;
    if (bottom < 0) lsrow = sfwk->h.height - 1 + bottom;
    nrows = lsrow - fsrow + 1;
    if (nrows <= 0) return;

    fdrow = 0;
    if (top > 0) fdrow = top;
    if (fdrow + nrows > dfwk->h.height) return;

    if (dfwk->h.proportional) {
        dfwk->gly[dglyph].width = fdbit + nbits;
        dfwk->gly[dglyph].bmprowsize = (dfwk->gly[dglyph].width + 7) / 8;
    }

    if (clear) {
        byxgly = ((dfwk->maxwidth + 7) / 8) * dfwk->h.height;
        memset(dfwk->gly[dglyph].bmp, 0, byxgly);
    }

    for (r=0; r<nrows; r++) {
        sp = sfwk->gly[sglyph].bmp + (sfwk->gly[sglyph].bmprowsize * (r + fsrow));
        dp = dfwk->gly[dglyph].bmp + (dfwk->gly[dglyph].bmprowsize * (r + fdrow));
        for (b=0; b<nbits; b++) {
            sbyte = (b + fsbit) / 8;
            sbit = 0x80 >> ((b + fsbit) % 8);
            if (sp[sbyte] & sbit) {
                dbyte = (b + fdbit) / 8;
                dbit = 0x80 >> ((b + fdbit) % 8);
                dp[dbyte] |= dbit;
            }
        }
    }
}
